/* Access points into the dynamically loaded Rx transport module
 *
 * Copyright (C) 2015 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version
 * 2 of the Licence, or (at your option) any later version.
 */

#ifndef _KAFS_TRANSPORT_H
#define _KAFS_TRANSPORT_H

#include <stdint.h>
#include <sys/socket.h>
#include <sys/uio.h>

struct kafs_transport_handle;
struct kafs_security_handle;
struct kafs_connection_handle;
struct kafs_call_handle;

enum kafs_connection_auth_level {
	kafs_connection_no_auth,
	kafs_connection_local_auth,
	kafs_connection_clear,
	kafs_connection_integrity_only,
	kafs_connection_encrypt,
};

enum kafs_recv_condition {
	kafs_recv_more,
	kafs_recv_final_ack,
	kafs_recv_complete,
	kafs_recv_complete_now_send,
	kafs_recv_abort,
	kafs_recv_net_error,
	kafs_recv_local_error,
	kafs_recv_busy,
};

#define kafs_transport_id "kAFS-utils tran"

struct kafs_transport_handle;
struct kafs_security_handle;
struct kafs_connection_handle;
struct kafs_service_handle;
struct kafs_call_handle;

typedef void (*kafs_incoming_call_func)(struct kafs_service_handle *service,
					void *caller_data,
					const struct sockaddr *peer,
					socklen_t *peerlen);

typedef void (*kafs_data_ready_func)(struct kafs_call_handle *call,
				     void *caller_data);

struct kafs_transport {
	char id[16];	/* = kafs_transport_id */
	char name[16];	/* Name of transport */

	struct kafs_transport_handle *(*init_transport)(const struct sockaddr *sa,
							socklen_t salen);

	void (*free_transport)(struct kafs_transport_handle *transport);

	struct kafs_security_handle *(*new_security)(struct kafs_transport_handle *transport,
						     const char *cell_name,
						     enum kafs_connection_auth_level level);

	void (*free_security)(struct kafs_security_handle *security);

	struct kafs_connection_handle *(*open_connection)(struct kafs_transport_handle *transport,
							  struct kafs_security_handle *security,
							  const struct sockaddr *sa,
							  socklen_t salen,
							  uint16_t service,
							  int exclusive_connection);

	void (*close_connection)(struct kafs_connection_handle *connection);

	struct kafs_call_handle *(*make_call)(struct kafs_connection_handle *connection,
					      kafs_data_ready_func data_ready_func,
					      void *caller_data);

	struct kafs_service_handle *(*new_service)(struct kafs_transport_handle *transport,
						   kafs_incoming_call_func new_call_func,
						   void *caller_data,
						   uint16_t local_port,
						   uint16_t local_service,
						   struct kafs_security_handle **permits);

	void (*shutdown_service)(struct kafs_service_handle *service);
	
	struct kafs_call_handle *(*accept_call)(struct kafs_service_handle *service,
						kafs_data_ready_func data_ready_func,
						void *caller_data,
						struct sockaddr *_sa,
						socklen_t *_salen);

	void (*abort_call)(struct kafs_call_handle *call,
			   uint32_t abort_code);

	void (*terminate_call)(struct kafs_call_handle *call);

	int (*send_data)(struct kafs_call_handle *call,
			 struct iovec *iov,
			 int ioc,
			 int more);

	int (*recv_data)(struct kafs_call_handle *call,
			 struct iovec *iov,
			 int ioc,
			 enum kafs_recv_condition *_cond,
			 uint32_t *_supplementary_info);

	int (*poll_connection)(struct kafs_connection_handle *conn,
			      int nowait);
};

extern struct kafs_transport kafs_transport;

#endif /* _KAFS_TRANSPORT_H */
